;------------------------------------------------------------
;
; MON386 - 80386 hardware debugging utility
; Copyright (c) 1992-2003,2004 by Alessandro Scotti
; http://www.ascotti.org/
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;
;
; Note: works in real mode only, special API such as those
; provided by DPMI or VCPI must be used in protected mode.
; May work in virtual (V86) mode if supported by DOS extender
; (most will).
;
; Note: if MON386 is used with EMM386 I have found that a
; code breakpoint set on the first instruction of an
; interrupt handler will not be catched. A data breakpoint
; on the interrupt table, however, will work.
;
;------------------------------------------------------------
LOCALS
JUMPS
INCLUDE MON386.INC

PUBLIC  MON3_Init
PUBLIC  MON3_Done
PUBLIC  MON3_SetBreakpoint
PUBLIC  MON3_SetSingleStep
PUBLIC  MON3_ClearBreakpoint
PUBLIC  MON3_ClearSingleStep

MON3_TEXT       SEGMENT USE16 WORD PUBLIC 'CODE'
                ASSUME  cs:MON3_TEXT, ds:NOTHING, es:NOTHING, ss:NOTHING

;------------------------------------------------------------
;
; Static variables
;
MON_OldInt01h   LABEL   DWORD           ; Old int 01h handler
MON_OldOfs01h   DW      0
MON_OldSeg01h   DW      0

MON_UsrInt01h   LABEL   DWORD           ; User break handler
MON_UsrOfs01h   DW      ?
MON_UsrSeg01h   DW      ?
MON_UsrStkPtr   LABEL   DWORD           ; User stack frame
MON_UsrStkOfs   DW      ?
MON_UsrStkSeg   DW      ?
MON_UsrDataSeg  DW      ?               ; User data segment

MON_Regs        REGISTERS       <>      ; Registers

MON_Flags       IFLAGS  <>

;============================================================
;
; Enables single-step interrupt
;
; Input:
;       none
; Output:
;       none
;
MON3_SetSingleStep      PROC    FAR
        or      [MON_Flags], MASK FL_STEP
        ret
MON3_SetSingleStep      ENDP

;============================================================
;
; Disables single-step interrupt
;
; Input:
;       none
; Output:
;       none
;
MON3_ClearSingleStep    PROC    FAR
        and     [MON_Flags], NOT MASK FL_STEP
        ret
MON3_ClearSingleStep    ENDP

;============================================================
;
; Removes a hardware breakpoint
;
; Input:
;       wBreakNumber    = breakpoint number (0,1,2,3)
; Output:
;       AX      = 0 on success, otherwise -1
;
.8086
MON3_ClearBreakpoint    PROC    FAR
        ARG     wBreakNumber:WORD       = ARG_SIZE
        .enter
        .push   ds, si, es, di          ; Save registers

        test    [MON_Flags], MASK FL_INIT
        jz      @@Exit                  ; Not intialized, exit
        mov     ax, -1                  ; Assume an error
        mov     bx, [wBreakNumber]
        cmp     bx, 04h                 ; Valid BP number?
        jae     @@Exit                  ; No, exit

.386P                                   ; Enable 386 privileged instructions
        mov     eax, dr7                ; Copy DR7 into EAX
        or      bx, bx
        jz      @@BP0
        dec     bx
        jz      @@BP1
        dec     bx
        jz      @@BP2
; Remove breakpoint #3
@@BP3:
        and     eax, NOT MASK G3        ; Global disable breakpoint
        jmp     @@SetDR7
@@BP2:
        and     eax, NOT MASK G2
        jmp     @@SetDR7
@@BP1:
        and     eax, NOT MASK G1
        jmp     @@SetDR7
@@BP0:
        and     eax, NOT MASK G0
        jmp     @@SetDR7
@@SetDR7:
        test    eax, MASK G0+MASK G1+MASK G2+MASK G3    ; All BPs disabled?
        jnz     @@Done                  ; No, update debug flags and exit
        and     eax, NOT MASK GEB       ; Reset global exact bit
@@Done:
        mov     dr7, eax                ; Copy EAX into DR7

        xor     ax, ax                  ; Return success
@@Exit:
.8086
        .pop    ds, si, es, di          ; Restore registers
        .leave
        ret     ARG_SIZE
MON3_ClearBreakpoint    ENDP

;============================================================
;
; Set hardware breakpoint
;
; Input:
;       vfpBreakAddress = breakpoint address
;       wBreakNumber    = breakpoint number (0,1,2,3)
;       wBreakType      = breakpoint type
; Output:
;       AX      = 0 on success, otherwise -1
;
; Breakpoint type format:
;
;       Bit 2-3 = BP length (0 = byte, 1 = word, 3 = dword)
;       Bit 0-1 = BP mode (0=code, 1 = write, 3 = read and write)
;
.8086
MON3_SetBreakpoint      PROC    FAR
        ARG     wBreakType:WORD,        \
                wBreakNumber:WORD,      \
                vfpBreakAddress:DWORD   = ARG_SIZE
        .enter
        .push   ds, si, es, di

        mov     cx, [wBreakType]
        mov     bx, [wBreakNumber]

        mov     ax, -1
        test    [MON_Flags], MASK FL_INIT
        jz      @@Exit                  ; Exit if not initialized
        cmp     bx, 04h
        jae     @@Exit                  ; Exit if invalid BP number

        test    cx, 00000011b           ; Test for data breakpoint
        jnz     @@1                     ; Continue if data BP
        xor     cx, cx                  ; Clear all bits to set code BP
@@1:

.386P                                   ; Enable 386 privileged instructions
        les     ax, [vfpBreakAddress]
        mov     dx, es                  ; Copy BP address into DX:AX
        movzx   eax, ax
        movzx   edx, dx
        shl     edx, 4
        add     edx, eax                ; Breakpoint linear address in EDX
        mov     eax, dr7                ; Copy DR7 into EAX
        or      bx, bx
        jz      @@BP0
        dec     bx
        jz      @@BP1
        dec     bx
        jz      @@BP2
; Set breakpoint #3
@@BP3:
        mov     dr3, edx                ; Load breakpoint address
        shl     ecx, BP3                ; Move bits to correct position
        and     ecx, MASK BP3           ; Mask off other bits
        and     eax, NOT (MASK BP3)     ; Mask off old bits
        or      eax, MASK G3            ; Global enable breakpoint
        jmp     @@SetBP
; Set breakpoint #2
@@BP2:
        mov     dr2, edx
        shl     ecx, BP2
        and     ecx, MASK BP2
        and     eax, NOT (MASK BP2)
        or      eax, MASK G2
        jmp     @@SetBP
; Set breakpoint #1
@@BP1:
        mov     dr1, edx
        shl     ecx, BP1
        and     ecx, MASK BP1
        and     eax, NOT (MASK BP1)
        or      eax, MASK G1
        jmp     @@SetBP
; Set breakpoint #0
@@BP0:
        mov     dr0, edx
        shl     ecx, BP0
        and     ecx, MASK BP0
        and     eax, NOT (MASK BP0)
        or      eax, MASK G0
@@SetBP:
        or      eax, ecx                ; Set breakpoint
        or      eax, MASK GEB           ; Set global exact bit
        mov     dr7, eax                ; Copy EAX into DR7

        xor     ax, ax                  ; Return success
@@Exit:
.8086
        .pop    ds, si, es, di
        .leave
        ret     ARG_SIZE
MON3_SetBreakpoint      ENDP

;============================================================
;
; Clear 386 debug registers
;
; Input:
;       none
; Output:
;       none
;
.386P                                   ; Enable 386 privileged instructions
SUB_ClearDebug  PROC    NEAR
        xor     eax, eax
        mov     dr7, eax
        mov     dr0, eax
        mov     dr1, eax
        mov     dr2, eax
        mov     dr3, eax
        mov     dr6, eax
        call    MON3_ClearSingleStep
        ret
SUB_ClearDebug  ENDP

;============================================================
;
; Terminate hardware debugging
;
; Input:
;       none
; Output:
;       AX      = 0 on success, otherwise -1
;
.8086
MON3_Done       PROC    FAR
        .push   ds, si, es, di

        mov     ax, -1                  ; Assume an error
        test    [MON_Flags], MASK FL_INIT
        jz      @@Exit                  ; Not initialized, exit

        and     [MON_Flags], NOT MASK FL_INIT
        call    SUB_ClearDebug          ; Disable hardware BPs
        les     bx, [MON_OldInt01h]
        xor     ax, ax
        mov     ds, ax
        cli
        mov     ds:[0004h], bx          ; Restore old interrupt 01h
        mov     ds:[0006h], es
        sti
        xor     ax, ax
@@Exit:
        .pop    ds, di, es, di
        ret
MON3_Done       ENDP

;============================================================
;
; Interrupt 01h handler
;
; Note:
;       this is an interrupt called whenever an enabled hardware
;       breakpoint or single-step occurs.
;
.386P
SUB_Int01h      PROC    FAR
        push    ax
        push    edx

        mov     edx, dr6                ; Copy debug control register into EDX
        xor     ax, ax
        test    edx, MASK BS            ; Single step?
        jnz     @@Catch                 ; Yes, get it
        bsf     ax, dx                  ; Search first bit set in DR6
        jz      @@Chain                 ; No bit set, exit
        cmp     al, 03h                 ; Is it a breakpoint bit?
        ja      @@Chain                 ; No, jump to old interrupt

; Check if it's a code breakpoint
        inc     ax
        push    ax
        shl     ax, 2                   ; Each break point is 4-bit long
        add     ax, 12                  ; Skip low word of EDX
        xchg    ax, cx                  ; Save CX and copy AX into CX
        mov     edx, dr7
        shr     edx, cl                 ; Move BP bits to bits 0-3 of DL
        test    dl, 0Fh                 ; Is any bit set?
        xchg    ax, cx                  ;   (restore CX)
        pop     ax                      ;   (restore AX)
        jnz     @@Catch                 ; Yes, it's a data BP
        or      [MON_Flags], MASK FL_RESUME

@@Catch:
        pop     edx
        jmp     SUB_BreakHandler

@@Chain:
        pop     edx
        pop     ax
        jmp     [MON_OldInt01h]
SUB_Int01h      ENDP

;============================================================
;
; Break handler
;
; Note: this function is called (usually by an interrupt handler)
;       with a breakpoint identifier in AX. On entry the stack
;       contains (in POP order) AX, IP, CS, FLAGS.
;       MON386 reserves some values for AX:
;               0       = single step,
;               1,2,3,4 = hardware breakpoint 0, 1, 2, 3.
;
.386P                                   ; Enable 386 privileged instructions
SUB_BreakHandler        PROC    FAR
        bts     [MON_Flags], FL_BREAK   ; Test and set nested call flag
        jnc     @@0
        pop     ax
        iret                            ; Exit if busy
@@0:
        sti                             ; Enable interrupts
        mov     [MON_Regs].rEAX, eax    ; Save used registers
        pop     [MON_Regs].wAX          ; Overwrite AX with value on stack
        mov     [MON_Regs].rEDX, edx
        pop     [MON_Regs].rIP          ; Get return address
        pop     [MON_Regs].rCS
        pop     [MON_Regs].rFLAGS

        push    ax
        mov     al, 20h
        out     20h, al
        pop     ax

; Setup register structure and stack frame
        mov     [MON_Regs].rEBX, ebx    ; Copy all registers to local structure
        mov     [MON_Regs].rECX, ecx
        mov     [MON_Regs].rEBP, ebp
        mov     [MON_Regs].rESP, esp
        mov     [MON_Regs].rESI, esi
        mov     [MON_Regs].rEDI, edi
        mov     [MON_Regs].rSS, ss
        mov     [MON_Regs].rDS, ds
        mov     [MON_Regs].rES, es
        mov     [MON_Regs].rFS, fs
        mov     [MON_Regs].rGS, gs
        pushfd
        pop     edx
        shr     edx, 16                 ; Get high word of extended flags
        mov     [MON_Regs].rEFLAGS, dx
        lss     sp, [MON_UsrStkPtr]     ; Switch to user stack

; Call user function
        push    ax                      ; Breakpoint number
        push    cs
        push    OFFSET MON_Regs         ; Pointer to register structure
        mov     ds, [MON_UsrDataSeg]    ; Setup DS
        call    [MON_UsrInt01h]         ; Call user breakpoint handler

; Restore registers and stack frame
        mov     ebx, [MON_Regs].rEBX   ; Get back registers
        mov     ecx, [MON_Regs].rECX
        mov     ebp, [MON_Regs].rEBP
        mov     esi, [MON_Regs].rESI
        mov     edi, [MON_Regs].rEDI
        mov     ds, [MON_Regs].rDS
        mov     es, [MON_Regs].rES
        mov     fs, [MON_Regs].rFS
        mov     gs, [MON_Regs].rGS
        cli                             ; Restore stack frame
        mov     ss, [MON_Regs].rSS
        mov     esp, [MON_Regs].rESP
        sti

; Enable trap bit of FLAGS if needed
        and     [MON_Regs].rFLAGS, NOT MASK fTRAP
        test    [MON_Flags], MASK FL_INIT
        jz      @@2                     ; Don't step if INIT flag clear
        test    [MON_Flags], MASK FL_STEP
        jz      @@2                     ; Don't step if STEP flag clear
        or      [MON_Regs].rFLAGS, MASK fTRAP
@@2:

; Enable resume bit of EFLAGS if needed
        mov     eax, dr6
        and     eax, NOT (MASK B0+MASK B1+MASK B2+MASK B3)
        mov     dr6, eax                ; Clear breakpoint bits
        btr     [MON_Flags], FL_RESUME  ; Resume from code breakpoint?
        jc      @@Resume                ; Yes, set RF bit in EFLAGS

; *** Exit point #1: back to caller
@@Exit:
        mov     eax, [MON_Regs].rEAX
        mov     edx, [MON_Regs].rEDX
        push    [MON_Regs].rFLAGS
        push    [MON_Regs].rCS
        push    [MON_Regs].rIP
        and     [MON_Flags], NOT MASK FL_BREAK
        iret

; *** Exit point #2: set resume flag and jump to caller
@@Resume:
        or      [MON_Regs].rEFLAGS, 0001h
        push    DWORD PTR [MON_Regs].rFLAGS
        movzx   eax, [MON_Regs].rCS
        push    eax
        movzx   eax, [MON_Regs].rIP             ; Convert IP to EIP
        push    eax
        mov     eax, [MON_Regs].rEAX
        mov     edx, [MON_Regs].rEDX
        and     [MON_Flags], NOT MASK FL_BREAK
        iretd
SUB_BreakHandler        ENDP

;============================================================
;
; Initialize hardware debugging
;
; Input:
;       vfpUsrBreak     = pointer to user breakpoint handler
;       vfpUsrStack     = pointer to user stack frame
; Output:
;       AX      = 0 on success, otherwise -1
;
; "C" prototype:
;       int     far pascal MON3_Init( void far *vfpUsrBreak );
; "Turbo Pascal" prototype:
;       function        MON3_Init( vfpUsrBreak: pointer ): integer; far;
;
.8086                                   ; Use only 8086 instructions here!
MON3_Init       PROC    FAR
        ARG     vfpUsrStack:DWORD,              \
                vfpUsrBreak:DWORD       = ARG_SIZE
        .enter
        .push   ds, si, es, di

        mov     ax, -1                  ; Assume an error
; Check hardware and software status
        test    [MON_Flags], MASK FL_INIT
        jnz     @@Exit                  ; Already initialized, exit
        call    SUB_TestCPU             ; Check if CPU is at least an 80386
        or      bx, bx                  ; Valid CPU?
        jnz     @@Exit                  ; No, exit
; Clear 386 hardware debug registers
        call    SUB_ClearDebug
; Install new int 01h
        .dos    3501h
        mov     [MON_OldOfs01h], bx
        mov     [MON_OldSeg01h], es
        les     bx, [vfpUsrBreak]
        mov     [MON_UsrOfs01h], bx
        mov     [MON_UsrSeg01h], es
        mov     [MON_UsrDataSeg], ds
        les     bx, [vfpUsrStack]
        and     bx, 0FFFEh              ; Make stack word aligned
        mov     [MON_UsrStkOfs], bx
        mov     [MON_UsrStkSeg], es
        xor     ax, ax
        mov     es, ax
        cli
        mov     WORD PTR es:[0004h], OFFSET SUB_Int01h
        mov     WORD PTR es:[0006h], SEG MON3_TEXT
        sti
        mov     [MON_Flags], MASK FL_INIT
        xor     ax, ax                  ; Return success
@@Exit:
        .pop    ds, si, es, di
        .leave
        ret     ARG_SIZE
MON3_Init       ENDP

;============================================================
;
; Check if CPU is at least an 80386
;
; Input:
;       none
; Output:
;       BX      = 0 on success, otherwise -1
;
.8086                                   ; Use only 8086 instructions here!
SUB_TestCPU     PROC    NEAR
        mov     bx, -1                  ; Assume an error
        pushf                           ; Save flags
        xor     ax, ax
        push    ax
        popf
        sti                             ; Keep interrupts enabled
        pushf
        pop     ax
        and     ax, 0F000h              ; Mask bits 12-15
        cmp     ax, 0F000h              ; Check if bits 12-15 are set
        je      @@Exit                  ; Yes, exit (8086 or 80186)
        mov     ax, 7000h               ; Now set bit 12-14 and reset bit 15
        push    ax                      ; (NT, IOPL)
        popf
        sti                             ; Keep interrupts enabled
        pushf
        pop     ax
        and     ax, 7000h               ; Check if pattern match
        je      @@Exit                  ; Yes, it is an 80286
        xor     bx, bx                  ; Ok, at least an 80386
@@Exit:
        popf                            ; Restore flags
        ret
SUB_TestCPU     ENDP

ENDS    MON3_TEXT
END
