(*
    MON386 - Test program
    Copyright (c) 1994-2003,2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
{$F+,R-,S-,X+}
program Mon3Test;
uses Crt, Dos, Mon386;

const
  STACKSIZE     = 4096;                 (* Interrupt stack size in words *)
type
  PByte = ^byte;
  D3_Control    = record
    Flags       : word;
    VSeg        : word;
    VOfs        : longint;
    COfs        : word;
    CSeg        : word;
    BOfs        : word;
    BSeg        : word;
    ListSize    : word;
    OName       : byte;
    OArg1       : byte;
    OArg2       : byte;
    OArg3       : byte;
    OPrefix     : word;
    Reserved    : array[ 1..10 ] of byte;
  end;
var
  Control       : D3_Control;
  Line          : string;
  StackFrame    : array[ 1..STACKSIZE ] of word;

function DIS3_Decode( P: pointer ): word; far; external;
{$L \DEV\DIS386\DIS386}

procedure EmuInt; assembler;
asm
  retf
end;

procedure Int60; assembler;
asm
  nop
  nop
  nop
  mov ax, 2
  iret
end;

procedure Dummy;
begin
  asm
    nop
    mov ax, 1
    mov ax, 2
    int $60
    nop
    mov cx, 4
    rep lodsb
  end;
end;

function Hex( W: word ): string;
const
  H = '0123456789ABCDEF';
var
  I: byte;
  S: string[4];
begin
  S := '';
  for I:=1 to 4 do begin
    S := Copy( H, W and $0F+1, 1) + S;
    W := W shr 4;
  end;
  Hex := S;
end;

function Decode: word;
begin
  Control.BOfs := Ofs( Line[1] );
  Control.BSeg := Seg( Line[1] );
  Decode := DIS3_Decode( @Control );
  Line[0] := #255;
  Line[0] := char(Pos( #0, Line )-1);
end;

procedure BreakHandler( N: word; R: PMon3Regs ); far;
label
  Entry, GetKey;
var
  C: char;
  P: PByte;
  Q: ^word;
  F: boolean;
begin
Entry:
  F := FALSE;
  Control.Flags := 0;
  Control.VSeg := R^.CS;
  Control.CSeg := R^.CS;
  Control.VOfs := R^.IP;
  Control.COfs := R^.IP;
  Control.ListSize := 23;
  Decode;
  if( N > 0 ) then
    WriteLn( 'Breakpoint #', N-1, '!' );
  MON3_SetSingleStep;
  while( Length(Line) < 40 ) do Line := Line+' ';
  Write( Line, Hex(R^.FLAGS), ':', Hex(R^.AX), ' [S]tep, [G]o: ' );
GetKey:
  case UpCase(ReadKey) of
    'G': MON3_ClearSingleStep;
    'C': begin
           MON3_ClearBreakpoint( 2 );
           goto GetKey;
         end;
    'A': begin
           MON3_Done;
           Halt;
         end;
    else begin
      P := R^.CSIP;
      if( P^ = $CD ) then begin
        (* Emulate INT instruction *)
        Inc( P );
        Dec( R^.SP, 6 );  (* Room for FLAGS, CS and IP *)
        Q := Ptr( R^.SS, R^.SP );
        Q^ := R^.IP+2;
        Inc( Q ); Q^ := R^.CS;
        Inc( Q ); Q^ := R^.FLAGS and $FEFF;
        R^.FLAGS := R^.FLAGS and $FCFF;  (* Emulate CLI and clear trap *)
        R^.IP := MemW[0:4*word(P^)];
        R^.CS := MemW[0:4*word(P^)+2];
        F := TRUE;
      end;
    end;
  end;
  WriteLn;
  if( F ) then goto Entry;
end;

var
  ICA: array[ 1..16 ] of byte;
  W: word;
  L: longint;

begin
  WriteLn( Hex(MemW[0:6]), ':', Hex(MemW[0:4]) );
  WriteLn( 'MON3TEST - Test program for MON386 (Turbo Pascal version)' );
  Move( Ptr( 0, $4F0 )^, ICA, 16 );

  if( MON3_Init( @BreakHandler, @StackFrame[STACKSIZE] ) <> 0 ) then begin
    WriteLn( 'Unable to initialize MON386.' );
    Halt;
  end;

  SetIntVec( $60, @Int60 );
  WriteLn( 'Installing code breakpoint 2 at @Dummy' );
  MON3_SetBreakpoint( @Dummy, 2, BP_CODE );

  WriteLn( 'Installing byte write breakpoint 3 at 0000:04F2h' );
(*  MON3_SetBreakpoint( Ptr( 0,$4F2 ), 3, BP_WBYTE ); *)

  WriteLn( 'Executing "Dummy"' );
  Dummy;

(*
  WriteLn( 'Reading from 0000:04F2h' );
  W := Mem[ 0:$4F2 ];

  WriteLn( 'Writing to   0000:04F2h' );
  Mem[ 0:$4F2 ] := 1;
*)

  MON3_Done;
  Move( ICA, Ptr( 0, $4F0 )^, 16 );
end.
